/* -*- C++ -*-
 * lemon/default_map.h - Part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2005 Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Research Group on Combinatorial Optimization, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_DEFAULT_MAP_H
#define LEMON_DEFAULT_MAP_H


#include <lemon/bits/array_map.h>
#include <lemon/bits/vector_map.h>

///\ingroup graphmapfactory
///\file
///\brief Graph maps that construct and destruct
///their elements dynamically.

namespace lemon {


  template <typename _Graph, typename _Item, typename _Value>
  struct DefaultMapSelector {
    typedef ArrayMap<_Graph, _Item, _Value> Map;
  };

  // bool
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, bool> {
    typedef VectorMap<_Graph, _Item, bool> Map;
  };

  // char
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, char> {
    typedef VectorMap<_Graph, _Item, char> Map;
  };

  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, signed char> {
    typedef VectorMap<_Graph, _Item, signed char> Map;
  };

  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, unsigned char> {
    typedef VectorMap<_Graph, _Item, unsigned char> Map;
  };


  // int
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, signed int> {
    typedef VectorMap<_Graph, _Item, signed int> Map;
  };

  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, unsigned int> {
    typedef VectorMap<_Graph, _Item, unsigned int> Map;
  };


  // short
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, signed short> {
    typedef VectorMap<_Graph, _Item, signed short> Map;
  };

  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, unsigned short> {
    typedef VectorMap<_Graph, _Item, unsigned short> Map;
  };


  // long
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, signed long> {
    typedef VectorMap<_Graph, _Item, signed long> Map;
  };

  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, unsigned long> {
    typedef VectorMap<_Graph, _Item, unsigned long> Map;
  };

  // \todo handling long long type


  // float
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, float> {
    typedef VectorMap<_Graph, _Item, float> Map;
  };


  // double
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, double> {
    typedef VectorMap<_Graph, _Item,  double> Map;
  };


  // long double
  template <typename _Graph, typename _Item>
  struct DefaultMapSelector<_Graph, _Item, long double> {
    typedef VectorMap<_Graph, _Item, long double> Map;
  };


  // pointer
  template <typename _Graph, typename _Item, typename _Ptr>
  struct DefaultMapSelector<_Graph, _Item, _Ptr*> {
    typedef VectorMap<_Graph, _Item, _Ptr*> Map;
  };

  /// \e
  template <
    typename _Graph, 
    typename _Item,
    typename _Value>
  class DefaultMap 
    : public DefaultMapSelector<_Graph, _Item, _Value>::Map {
  public:
    typedef typename DefaultMapSelector<_Graph, _Item, _Value>::Map Parent;
    typedef DefaultMap<_Graph, _Item, _Value> Map;
    
    typedef typename Parent::Graph Graph;
    typedef typename Parent::Value Value;

    DefaultMap(const Graph& _g) : Parent(_g) {}
    DefaultMap(const Graph& _g, const Value& _v) : Parent(_g, _v) {}

  };


  /// \e
  template <typename _Base> 
  class MappableGraphExtender : public _Base {
  public:

    typedef MappableGraphExtender<_Base> Graph;
    typedef _Base Parent;

    typedef typename Parent::Node Node;
    typedef typename Parent::NodeIt NodeIt;

    typedef typename Parent::Edge Edge;
    typedef typename Parent::EdgeIt EdgeIt;

    
    template <typename _Value>
    class NodeMap 
      : public IterableMapExtender<DefaultMap<Graph, Node, _Value> > {
    public:
      typedef MappableGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, Node, _Value> > Parent;

      NodeMap(const Graph& _g) 
	: Parent(_g) {}
      NodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      NodeMap& operator=(const NodeMap& cmap) {
	return operator=<NodeMap>(cmap);
      }


      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concecpt and could be indiced by the current item set of
      /// the NodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      NodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Node, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Node it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }

    };

    template <typename _Value>
    class EdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, Edge, _Value> > {
    public:
      typedef MappableGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, Edge, _Value> > Parent;

      EdgeMap(const Graph& _g) 
	: Parent(_g) {}
      EdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      EdgeMap& operator=(const EdgeMap& cmap) {
	return operator=<EdgeMap>(cmap);
      }

      template <typename CMap>
      EdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Edge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Edge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };
    
  };

  /// \e
  template <typename _Base> 
  class MappableEdgeSetExtender : public _Base {
  public:

    typedef MappableEdgeSetExtender<_Base> Graph;
    typedef _Base Parent;

    typedef typename Parent::Edge Edge;
    typedef typename Parent::EdgeIt EdgeIt;

    template <typename _Value>
    class EdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, Edge, _Value> > {
    public:
      typedef MappableEdgeSetExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, Edge, _Value> > Parent;

      EdgeMap(const Graph& _g) 
	: Parent(_g) {}
      EdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      EdgeMap& operator=(const EdgeMap& cmap) {
	return operator=<EdgeMap>(cmap);
      }

      template <typename CMap>
      EdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Edge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Edge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };
    
  };

  /// \e
  template <typename _Base> 
  class MappableUndirGraphExtender : 
    public MappableGraphExtender<_Base> {
  public:

    typedef MappableUndirGraphExtender Graph;
    typedef MappableGraphExtender<_Base> Parent;

    typedef typename Parent::UndirEdge UndirEdge;

    template <typename _Value>
    class UndirEdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, UndirEdge, _Value> > {
    public:
      typedef MappableUndirGraphExtender Graph;
      typedef IterableMapExtender<
	DefaultMap<Graph, UndirEdge, _Value> > Parent;

      UndirEdgeMap(const Graph& _g) 
	: Parent(_g) {}
      UndirEdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      UndirEdgeMap& operator=(const UndirEdgeMap& cmap) {
	return operator=<UndirEdgeMap>(cmap);
      }

      template <typename CMap>
      UndirEdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UndirEdge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UndirEdge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };


  };

  /// \e
  template <typename _Base> 
  class MappableUndirEdgeSetExtender : 
    public MappableEdgeSetExtender<_Base> {
  public:

    typedef MappableUndirEdgeSetExtender Graph;
    typedef MappableEdgeSetExtender<_Base> Parent;

    typedef typename Parent::UndirEdge UndirEdge;

    template <typename _Value>
    class UndirEdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, UndirEdge, _Value> > {
    public:
      typedef MappableUndirEdgeSetExtender Graph;
      typedef IterableMapExtender<
	DefaultMap<Graph, UndirEdge, _Value> > Parent;

      UndirEdgeMap(const Graph& _g) 
	: Parent(_g) {}
      UndirEdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}

      UndirEdgeMap& operator=(const UndirEdgeMap& cmap) {
	return operator=<UndirEdgeMap>(cmap);
      }

      template <typename CMap>
      UndirEdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UndirEdge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UndirEdge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };


  };


  template <typename _Base>
  class MappableUndirBipartiteGraphExtender : public _Base {
  public:

    typedef _Base Parent;
    typedef MappableUndirBipartiteGraphExtender Graph;

    typedef typename Parent::Node Node;
    typedef typename Parent::UpperNode UpperNode;
    typedef typename Parent::LowerNode LowerNode;
    typedef typename Parent::Edge Edge;
    typedef typename Parent::UndirEdge UndirEdge;
    
    template <typename _Value>
    class UpperNodeMap 
      : public IterableMapExtender<DefaultMap<Graph, UpperNode, _Value> > {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, UpperNode, _Value> > 
      Parent;
    
      UpperNodeMap(const Graph& _g) 
	: Parent(_g) {}
      UpperNodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      UpperNodeMap& operator=(const UpperNodeMap& cmap) {
	return operator=<UpperNodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the UpperNodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      UpperNodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UpperNode, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UpperNode it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };

    template <typename _Value>
    class LowerNodeMap 
      : public IterableMapExtender<DefaultMap<Graph, LowerNode, _Value> > {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, LowerNode, _Value> > 
      Parent;
    
      LowerNodeMap(const Graph& _g) 
	: Parent(_g) {}
      LowerNodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      LowerNodeMap& operator=(const LowerNodeMap& cmap) {
	return operator=<LowerNodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the LowerNodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      LowerNodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<LowerNode, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	LowerNode it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };

  protected:

    template <typename _Value>
    class NodeMapBase : public Parent::NodeNotifier::ObserverBase {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;

      typedef Node Key;
      typedef _Value Value;

      /// The reference type of the map;
      typedef typename LowerNodeMap<_Value>::Reference Reference;
      /// The pointer type of the map;
      typedef typename LowerNodeMap<_Value>::Pointer Pointer;
      
      /// The const value type of the map.
      typedef const Value ConstValue;
      /// The const reference type of the map;
      typedef typename LowerNodeMap<_Value>::ConstReference ConstReference;
      /// The pointer type of the map;
      typedef typename LowerNodeMap<_Value>::ConstPointer ConstPointer;

      typedef True ReferenceMapTag;

      NodeMapBase(const Graph& _g) 
	: graph(&_g), lowerMap(_g), upperMap(_g) {
	Parent::NodeNotifier::ObserverBase::attach(_g.getNotifier(Node()));
      }
      NodeMapBase(const Graph& _g, const _Value& _v) 
	: graph(&_g), lowerMap(_g, _v), 
	  upperMap(_g, _v) {
	Parent::NodeNotifier::ObserverBase::attach(_g.getNotifier(Node()));
      }

      virtual ~NodeMapBase() {      
	if (Parent::NodeNotifier::ObserverBase::attached()) {
	  Parent::NodeNotifier::ObserverBase::detach();
	}
      }
    
      ConstReference operator[](const Key& node) const {
	if (Parent::upper(node)) {
	  return upperMap[node];
	} else {
	  return lowerMap[node];
	}
      } 

      Reference operator[](const Key& node) {
	if (Parent::upper(node)) {
	  return upperMap[node];
	} else {
	  return lowerMap[node];
	}
      }

      void set(const Key& node, const Value& value) {
	if (Parent::upper(node)) {
	  upperMap.set(node, value);
	} else {
	  lowerMap.set(node, value);
	}
      }

    protected:
      
      virtual void add(const Node&) {}
      virtual void erase(const Node&) {}
      virtual void clear() {}
      virtual void build() {}

      const Graph* getGraph() const { return graph; }
      
    private:
      const Graph* graph;
      LowerNodeMap<_Value> lowerMap;
      UpperNodeMap<_Value> upperMap;
    };
    
  public:

    template <typename _Value>
    class NodeMap 
      : public IterableMapExtender<NodeMapBase<_Value> > {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;
      typedef IterableMapExtender< NodeMapBase<_Value> > Parent;
    
      NodeMap(const Graph& _g) 
	: Parent(_g) {}
      NodeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      NodeMap& operator=(const NodeMap& cmap) {
	return operator=<NodeMap>(cmap);
      }
    

      /// \brief Template assign operator.
      ///
      /// The given parameter should be conform to the ReadMap
      /// concept and could be indiced by the current item set of
      /// the NodeMap. In this case the value for each item
      /// is assigned by the value of the given ReadMap. 
      template <typename CMap>
      NodeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Node, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Node it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    
    };



    template <typename _Value>
    class EdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, Edge, _Value> > {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, Edge, _Value> > Parent;
    
      EdgeMap(const Graph& _g) 
	: Parent(_g) {}
      EdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      EdgeMap& operator=(const EdgeMap& cmap) {
	return operator=<EdgeMap>(cmap);
      }
    
      template <typename CMap>
      EdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<Edge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	Edge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };

    template <typename _Value>
    class UndirEdgeMap 
      : public IterableMapExtender<DefaultMap<Graph, UndirEdge, _Value> > {
    public:
      typedef MappableUndirBipartiteGraphExtender Graph;
      typedef IterableMapExtender<DefaultMap<Graph, UndirEdge, _Value> > 
      Parent;
    
      UndirEdgeMap(const Graph& _g) 
	: Parent(_g) {}
      UndirEdgeMap(const Graph& _g, const _Value& _v) 
	: Parent(_g, _v) {}
    
      UndirEdgeMap& operator=(const UndirEdgeMap& cmap) {
	return operator=<UndirEdgeMap>(cmap);
      }
    
      template <typename CMap>
      UndirEdgeMap& operator=(const CMap& cmap) {
	checkConcept<concept::ReadMap<UndirEdge, _Value>, CMap>();
	const typename Parent::Graph* graph = Parent::getGraph();
	UndirEdge it;
	for (graph->first(it); it != INVALID; graph->next(it)) {
	  Parent::set(it, cmap[it]);
	}
	return *this;
      }
    };
  
  };

}

#endif
